"""Top-level package for Cfg4Py."""
from cfg4py.core import (
    RedisConfigFetcher,
    RemoteConfigFetcher,
    _cfg_obj,
    config_remote_fetcher,
    config_server_role,
    enable_logging,
    envar,
    get_config_dir,
    init,
    update_config,
)

__author__ = """Aaron Yang"""
__email__ = "code@jieyu.ai"
__version__ = "0.9.0"


def get_instance():
    return _cfg_obj


__all__ = [
    "RemoteConfigFetcher",
    "enable_logging",
    "config_remote_fetcher",
    "init",
    "update_config",
    "RedisConfigFetcher",
    "config_server_role",
    "envar",
    "get_instance",
    "get_config_dir",
]



"""Main module."""
import logging.config
import os
import re
from collections.abc import Mapping
from io import StringIO

from watchdog.events import FileModifiedEvent, FileSystemEventHandler
from watchdog.observers import Observer

from apscheduler.schedulers.background import BackgroundScheduler
from cfg4py.config import Config
from ruamel.yaml import YAML
from ruamel.yaml.error import YAMLError

logger = logging.getLogger(__name__)

envar = "__cfg4py_server_role__"
_scheduler = BackgroundScheduler()
_remote_fetcher = None
_dump_on_change: bool = False

yaml = YAML(typ="safe")

# handle local configuration file change
_local_observer = None
_cfg_obj = Config()

_cfg_local = {}
_cfg_remote = {}
_strict = True

_local_config_dir: str = ""


class RemoteConfigFetcher:
    def fetch(self) -> str:
        raise NotImplementedError("sub class must implement this!")  # pragma: no cover


class RedisConfigFetcher(RemoteConfigFetcher):
    def __init__(
        self, key: str, host: str = "localhost", port: int = 6379, db: int = 0, **kwargs
    ):
        self.key = key
        from redis import StrictRedis  # type: ignore

        self.client = StrictRedis(
            host, port=port, db=db, decode_responses="utf-8", **kwargs
        )

    def fetch(self) -> dict:
        settings = None
        try:
            logger.info("fetching configuration from redis server")
            settings = self.client.get(self.key)
            return _load_and_replace_envar(settings)
        except Exception as e:  # pragma: no cover
            logger.exception(e)

        return {}


class LocalConfigChangeHandler(FileSystemEventHandler):
    def dispatch(self, event):
        if isinstance(event, FileModifiedEvent):
            _load_from_local_file()


def _mixin(d, u):
    # noinspection SpellCheckingInspection
    """
    if  value x in "keyx:valuex" pair is list, it's will be replaced
    :param d:
    :param u:
    :return:
    """
    u = u or {}
    d = d or {}
    for k, v in u.items():
        if isinstance(v, Mapping):
            d[k] = _mixin(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def _to_obj(obj, conf: dict):
    for key, value in conf.items():
        if type(value) == dict:
            _obj = Config()
            setattr(obj, key, _obj)
            _to_obj(_obj, value)
        else:
            setattr(obj, key, value)


def _refresh():
    global _cfg_local, _cfg_remote
    _cfg_remote = _remote_fetcher.fetch()
    merged = _mixin(_cfg_remote, _cfg_local)
    update_config(merged)


def enable_logging(level=logging.INFO, log_file=None, file_size=10, file_count=7):
    """
    Enable basic log function for the application

    if log_file is None, then it'll provide console logging, otherwise, the console
    logging is turned off, all events will be logged into the provided file.

    Args:
        level: the log level, one of logging.DEBUG, logging.INFO, logging.WARNING,
        logging.Error
        log_file: the absolute file path for the log.
        file_size: file size in MB unit
        file_count: how many backup files leaved in disk

    Returns:
        None
    """

    assert file_count > 0
    assert file_size > 0

    from logging import handlers

    formatter = logging.Formatter(
        "%(asctime)s %(levelname)-1.1s %(filename)s:%(lineno)s | %(message)s"
    )

    _logger = logging.getLogger()
    _logger.setLevel(level)

    if log_file is None:
        console = logging.StreamHandler()
        console.setFormatter(formatter)
        _logger.addHandler(console)
    else:
        file_dir = os.path.dirname(log_file)
        os.makedirs(file_dir, exist_ok=True)
        rotating_file = handlers.RotatingFileHandler(
            log_file, maxBytes=1024 * 1024 * file_size, backupCount=file_count
        )
        rotating_file.setFormatter(formatter)
        _logger.addHandler(rotating_file)


def config_remote_fetcher(fetcher: RemoteConfigFetcher, interval: int = 300):
    """
    config a remote configuration fetcher, which will pull the settings on every
     `refresh_interval`
    Args:
        fetcher: sub class of `RemoteConfigFetcher`
        interval: how long should cfg4py to pull the configuration from remote

    Returns:

    """
    global _remote_fetcher
    _remote_fetcher = fetcher

    _scheduler.add_job(_refresh, "interval", seconds=interval)
    _scheduler.start()


def build(save_to: str):
    global _cfg_obj
    with open(os.path.join(os.path.dirname(__file__), "config.py"), "r") as origin:
        lines = origin.readlines()
        lines.append("\n")

    no_instance = [
        f"{' ' * 4}def __init__(self):\n",
        f"{' ' * 8}raise TypeError(\"Do NOT instantiate this class\")\n",
    ]
    lines.extend(no_instance)
    with open(save_to, encoding="utf-8", mode="w") as f:
        lines = _schema_from_obj_(_cfg_obj, lines)
        f.writelines("".join(lines))


def _schema_from_obj_(obj, lines, depth: int = 0):
    """
    build a python file for auto-complete.
    """
    depth += 1
    if isinstance(obj, Config):
        for name in obj.__dict__.copy().keys():
            if name.startswith("__"):
                continue
            child = getattr(obj, name)
            if callable(child):
                continue

            if isinstance(child, Config):
                lines.append("\n")
                lines.append(f"{' ' * 4 * depth}class {name}:\n")
                _schema_from_obj_(child, lines, depth)
            else:
                _type = f"{type(child)}"
                _type = re.sub(r".*\'(.*)\'>", r"\1", _type)
                if _type != "NoneType":
                    lines.append(f"{' ' * 4 * depth}{name}: Optional[{_type}] = None\n")
                else:
                    lines.append(f"{' ' * 4 * depth}{name} = None\n")
    else:
        print(obj)

    return lines


def init(local_cfg_path: str = None, dump_on_change=True, strict=False):
    """
    create cfg object.
    Args:
        local_cfg_path: the directory name where your configuration files exist
        dump_on_change: if configuration is updated, whether or not to dump them into
         log file

    Returns:
    """
    global _local_config_dir, _dump_on_change, _remote_fetcher, _local_observer
    global _cfg_obj, _cfg_local, _cfg_remote
    global _strict

    _strict = strict
    _dump_on_change = dump_on_change
    if local_cfg_path:
        _local_config_dir = os.path.expanduser(local_cfg_path)

        # handle local configuration file change
        _local_observer = Observer()
        _local_observer.schedule(
            LocalConfigChangeHandler(), _local_config_dir, recursive=False
        )
        _local_observer.start()

        _cfg_local = _load_from_local_file()
        update_config(_mixin(_cfg_remote, _cfg_local))

        # todo: will this overwrite existing file occasionally?
        save_to = os.path.join(_local_config_dir, "schema.py")
        build(save_to)
    return _cfg_obj


def yaml_dump(conf, options=None):
    if options is None:
        options = {}
    string_stream = StringIO()
    try:
        yaml.dump(conf, string_stream, **options)
        output_str = string_stream.getvalue()
    finally:
        string_stream.close()
    return output_str


def update_config(conf: dict):
    global _cfg_obj

    if "logging" in conf:
        _process_logging_settings(conf["logging"])

        del conf["logging"]

    if _dump_on_change:
        logger.info("configuration is\n%s", yaml_dump(conf))

    _to_obj(_cfg_obj, conf)
    return _cfg_obj


def _process_logging_settings(conf: dict):
    logging.config.dictConfig(conf)


def _guess_extension():
    files = os.listdir(_local_config_dir)
    counter = {".yml": 0, ".yaml": 0}

    for f in files:
        _, ext = os.path.splitext(f)
        if ext == ".yml":
            counter[ext] += 1
        elif ext == ".yaml":
            counter[ext] += 1

    _max = 0
    ext = ""
    for key in counter.keys():
        if counter[key] > _max:
            _max = counter[key]
            ext = key

    if ext in [".yml", ".yaml"]:
        # noinspection PyPackageRequirements
        return ext
    else:
        msg = "No config files present, or file format is not yaml."
        raise FileNotFoundError(msg)


def get_config_dir():
    return _local_config_dir


def _load_and_replace_envar(content: str):
    """parse content, replace placeholder with environment variables, and load with yaml

    Args:
        content (str): the content of configurations
    """
    pattern = re.compile(r".*?\${(\w+)}.*?")
    match = pattern.findall(content)
    if match:
        replaced = content
        for g in match:
            replaced = replaced.replace(
                f"${{{g}}}", os.environ.get(g, f"ERROR_ENVAR_NOT_SET[{g}]")
            )
        try:
            return yaml.load(replaced)
        except YAMLError as e:
            logger.error("failed to parse:%s\n", content)
            raise e

    try:
        return yaml.load(content)
    except YAMLError as e:
        logger.error("failed to parse：%s\n", content)
        raise e


def _load_from_local_file() -> dict:
    """
    read configuration hierarchically from disk
    Args:

    Returns:

    """
    global _strict

    conf = {}

    role = os.getenv(envar, "")
    if role == "" and _strict:
        msg = f"You must config environment variables {envar} as one of"
        "'DEV, TEST, PRODUCTION'"
        raise EnvironmentError(msg)
    try:
        ext = _guess_extension()
        with open(
            os.path.join(_local_config_dir, f"defaults{ext}"), "r", encoding="utf-8"
        ) as base:
            conf = _load_and_replace_envar(base.read(-1))

        if role == "PRODUCTION":
            with open(
                os.path.join(_local_config_dir, f"production{ext}"),
                "r",
                encoding="utf-8",
            ) as prod:
                _prod = _load_and_replace_envar(prod.read(-1))
                _mixin(conf, _prod)

        elif role == "TEST":
            with open(
                os.path.join(_local_config_dir, f"test{ext}"), "r", encoding="utf-8"
            ) as test:
                _test = _load_and_replace_envar(test.read(-1))
                _mixin(conf, _test)
        elif role == "DEV":
            with open(
                os.path.join(_local_config_dir, f"dev{ext}"), "r", encoding="utf-8"
            ) as dev:
                _dev = _load_and_replace_envar(dev.read(-1))
                _mixin(conf, _dev)
        else:
            pass
    except FileNotFoundError as e:
        if e.filename.find("defaults") != -1:
            raise FileNotFoundError("Failed to find default configuration file")
    except Exception as e:
        logger.exception(e)

    return conf


# noinspection PyUnusedLocal
def config_server_role(role: str):
    os.environ[envar] = role



# noqa
from typing import Optional


class Config(object):
    __access_counter__ = 0

    def __cfg4py_reset_access_counter__(self):
        self.__access_counter__ = 0

    def __getattribute__(self, name):
        obj = object.__getattribute__(self, name)
        if name.startswith("__") and name.endswith("__"):
            return obj

        if callable(obj):
            return obj

        self.__access_counter__ += 1
        return obj



#!/usr/bin/env python

"""The setup script."""

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "watchdog>=0.9.0",
    "apscheduler>=3.6.3",
    "fire>=0.2.1",
    "ruamel.yaml>=0.16.0",
]

setup_requirements = []

setup(
    author="Aaron Yang",
    author_email="aaron_yang@jieyu.ai",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Easy config module supports cascading design, adpative deployment, auto code complete and more",
    install_requires=requirements,
    license="BSD license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="cfg4py",
    name="cfg4py",
    packages=find_packages(include=["cfg4py", "cfg4py.*"]),
    setup_requires=setup_requirements,
    url="https://github.com/zillionare/cfg4py",
    version="0.9.0",
    zip_safe=False,
    entry_points={"console_scripts": ["cfg4py=cfg4py.command_line:main"]},
)



import logging
import os
import sys
from typing import Optional

import fire
from cfg4py import enable_logging, envar, init
from ruamel.yaml import YAML

enable_logging()


class Command:
    def __init__(self):
        self.resource_path = os.path.normpath(
            os.path.join(os.path.dirname(__file__), "resources/")
        )
        self.yaml = YAML(typ="safe")  # default, if not specfied, is 'rt' (round-trip)
        self.yaml.default_flow_style = False

        with open(
            os.path.join(self.resource_path, "template.yaml"), "r", encoding="utf-8"
        ) as f:
            self.templates = self.yaml.load(f)

        self.transformed = self._transform()

    def build(self, config_dir: str):
        """Compile configuration files into a python class file, which is used by IDE's
        auto-complete function

        Args:
            config_dir: The folder where your configuration files located

        Returns:

        """
        if not os.path.exists(config_dir):
            print(f"path {config_dir} not exists")
            sys.exit(-1)

        count = 0
        for f in os.listdir(config_dir):
            if (
                f.startswith("default")
                or f.startswith("dev")
                or f.startswith("test")
                or f.startswith("production")
            ):
                print(f"found {f}")
                count += 1

        if count > 0:
            print(f"{count} files found in total")
        else:
            print("the folder contains no valid configuration files")
            sys.exit(-1)

        try:
            init(config_dir)
            sys.path.insert(0, config_dir)
            from schema import Config  # type: ignore # noqa

            output_file = f"{os.path.join(config_dir, 'schema')}"
            msg = f"Config file is built with success and saved at {output_file}"
            print(msg)
        except Exception as e:  # pragma: no cover
            logging.exception(e)
            print("Config file built failure.")

    def _choose_dest_dir(self, dst):
        if dst is None:
            dst = input("Where should I save configuration files?\n")

        if os.path.exists(dst):
            for f in os.listdir(dst):
                msg = f"The folder already contains {f}, please choose clean one."

                if f in ["defaults.yaml", "dev.yaml", "test.yaml", "production.yaml"]:
                    print(msg)
                    return None
            return dst
        else:
            create = input("Path not exists, create('Q' to exit)? [Y/n]")
            if create.upper() == "Y":
                os.makedirs(dst, exist_ok=True)
                return dst
            elif create.upper() == "Q":
                sys.exit(-1)
            else:
                return None

    def scaffold(self, dst: Optional[str]):
        """Creates initial configuration files based on our choices.
        Args:
            dst:

        Returns:

        """
        print("Creating a configuration boilerplate:")
        dst = self._choose_dest_dir(dst)
        while dst is None:
            dst = self._choose_dest_dir(dst)

        yaml = YAML(typ="safe")  # default, if not specfied, is 'rt' (round-trip)
        with open(
            os.path.join(self.resource_path, "template.yaml"), "r", encoding="utf-8"
        ) as f:
            templates = yaml.load(f)

        print("Which flavors do you want?")
        print("-" * 20)
        prompt = """
        0  - console + rotating file logging
        10 - redis/redis-py (gh://andymccurdy/redis-py)
        11 - redis/aioredis (gh://aio-libs/aioredis)
        20 - mysql/PyMySQL (gh://PyMySQL/PyMySQL)
        30 - postgres/asyncpg (gh://MagicStack/asyncpg)
        31 - postgres/psycopg2 (gh://psycopg/psycopg2)
        40 - mq/pika (gh://pika/pika)
        50 - mongodb/pymongo (gh://mongodb/mongo-python-driver)
        """
        print(prompt)
        chooses = input(
            "Please choose flavors by index, separated each by a comma(,):\n"
        )
        flavors = {}
        mapping = {
            "0": "logging",
            "1": "redis",
            "2": "mysql",
            "3": "postgres",
            "4": "mq",
            "5": "mongodb",
        }
        for index in chooses.strip(" ").split(","):
            if index == "0":
                flavors["logging"] = templates["logging"]
                continue

            try:
                major = mapping[index[0]]
                minor = int(index[1])
                flavors[major] = list(templates[major][minor].values())[0]
            except (ValueError, KeyError):
                print(f"Wrong index {index}, skipped.")
                continue

        with open(os.path.join(dst, "defaults.yaml"), "w", encoding="utf-8") as f:
            f.writelines(
                "#auto generated by Cfg4Py: https://github.com/jieyu-tech/cfg4py\n"
            )
            yaml.dump(flavors, f)

        print(f"Cfg4Py has generated the following files under {dst}:")
        print("defaults.yaml")
        for name in ["dev.yaml", "test.yaml", "production.yaml"]:
            with open(os.path.join(dst, name), "w", encoding="utf8") as f:
                f.writelines(
                    "#auto generated by Cfg4Py: https://github.com/jieyu-tech/cfg4py\n"
                )
                print(name)

        with open(os.path.join(dst, "defaults.yaml"), "r", encoding="utf-8") as f:
            print("Content in defaults.yaml")
            for line in f.readlines():
                print(line.replace("\n", ""))

    def _show_supported_config(self):
        print("Support the following configurations:")
        for key in self.templates.keys():
            item = self.templates.get(key)
            if isinstance(item, dict):
                print(f"  {key}")
            elif isinstance(item, list):
                sub_keys = []
                for sub_item in item:
                    sub_keys.append(f"{key}/{list(sub_item.keys())[0]}")
                print(f"  {key}: {', '.join(sub_keys)}")

    def _transform(self):
        transformed = {}
        for key in self.templates:
            if isinstance(self.templates[key], dict):
                transformed[key] = self.templates[key]
            elif isinstance(self.templates[key], list):
                for item in self.templates[key]:
                    item_key = list(item.keys())[0]
                    transformed[f"{key}/{item_key}"] = item

        return transformed

    def hint(self, what: str = None, usage: bool = False):
        """show a cheat sheet for configurations.
        for example:
            cfg4py hint mysql
        this will print how to configure PyMySQL
        :param what
        :param usage
        """

        if what is None or (
            (what not in self.templates) and what not in self.transformed
        ):
            return self._show_supported_config()

        usage_key = f"{what}_usage"

        if usage_key in self.templates and usage:
            print("Usage:", self.templates.get(usage_key))

        if what in self.templates:
            self.yaml.dump(self.templates[what], sys.stdout)

        if usage_key in self.transformed and usage:
            print("Usage:", self.transformed.get(usage_key))

        if what in self.transformed:
            self.yaml.dump(self.transformed[what], sys.stdout)

    def set_server_role(self):
        print("please add the following line into your .bashrc:\n")
        print(f"export {envar}=DEV\n")
        msg = "You need to change DEV to TEST | PRODUCTION according to its actual role\
         accordingly"
        print(msg)


def main():
    cmd = Command()  # pragma: no cover
    fire.Fire(
        {  # pragma: no cover
            "build": cmd.build,
            "scaffold": cmd.scaffold,
            "hint": cmd.hint,
            "set_server_role": cmd.set_server_role,
        }
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Tests for `cfg4py` package."""
import logging
import os
import shutil
import sys
import tempfile
import time
import unittest
import unittest.mock

import cfg4py
from cfg4py import core
from cfg4py.command_line import Command

logger = logging.getLogger(__name__)


def early_jump(msg):
    logger.info("Mock exit: %s", msg)
    raise SystemExit


class TestCfg4Py(unittest.TestCase):
    """Tests for `cfg4py` package."""

    def setUp(self):
        """Set up test fixtures, if any."""
        cfg4py.enable_logging()
        os.environ["__cfg4py_server_role__"] = "TEST"
        self.resource_path = os.path.join(
            os.path.dirname(__file__), "../cfg4py/resources/"
        )
        self.resource_path = os.path.normpath(self.resource_path)
        self.home = os.path.normpath(os.path.join(os.path.dirname(__file__), ".."))
        self.output = os.path.join(tempfile.gettempdir(), "cfg4py_test")
        shutil.rmtree(self.output, ignore_errors=True)

    def tearDown(self):
        """Tear down test fixtures, if any."""

    def test_000_create_config(self):
        """Test something."""
        logger.info("resource_path is %s", self.resource_path)
        cfg = cfg4py.init(self.resource_path, strict=True)
        self.assertEqual(cfg.services.redis.host, "127.0.0.1")

    def test_001_update_config(self):
        conf = {
            "aaron": {"surname": "yang"},
            "services": {"redis": {"host": "localhost"}},
        }

        cfg = cfg4py.update_config(conf)
        self.assertEqual("localhost", cfg.services.redis.host)

    def test_001_build(self):
        conf = {
            "aaron": {"surname": "yang"},
            "services": {"redis": {"host": "localhost"}},
        }

        cfg4py.update_config(conf)
        core.build(os.path.join(self.resource_path, "schema.py"))
        try:
            sys.path.insert(0, self.resource_path)
            from schema import Config  # noqa

            # no exception means the file has been generated successfully
            self.assertTrue(True)
        except Exception:
            self.assertTrue(False)

    def test_002_create_config(self):
        from cfg4py.resources.schema import Config

        os.environ["__cfg4py_server_role__"] = "DEV"
        cfg: Config = cfg4py.init(self.resource_path, strict=True)
        print("cfg.services.redis.host is", cfg.services.redis.host)

    def test_003_config_remote_fetcher(self):
        import json

        from redis import StrictRedis  # type: ignore

        from cfg4py import RedisConfigFetcher

        r = StrictRedis("localhost")
        r.set(
            "my_app_config",
            json.dumps(
                {
                    "services": {
                        "redis": {
                            "host": "192.168.3.1",
                        },
                        "redis2": {"host": "192.168.3.2"},
                    }
                }
            ),
        )

        cfg = cfg4py.init(self.resource_path, strict=True)
        fetcher = RedisConfigFetcher(key="my_app_config")
        logger.info("configuring a remote fetcher")
        cfg4py.config_remote_fetcher(fetcher, 0.1)
        time.sleep(0.5)
        logger.info(
            "please check log output to see if update_config is called at least 1 once"
        )
        # the one from dev.yaml
        self.assertEqual(cfg.services.redis.host, "127.0.0.1")
        # the one from redis
        self.assertEqual(cfg.services.redis2.host, "192.168.3.2")

    def test_004_command_hint(self):
        cmd = Command()
        logger.info("please check the output")
        cmd.hint("pip")
        cmd.hint("pip", True)
        cmd.hint("postgres/psycopg2")
        cmd.hint("postgres/psycopg2", True)
        cmd.hint("pip/tshinghua")

    def test_005_sigleton(self):
        cfg1 = cfg4py.get_instance()
        cfg2 = cfg4py.get_instance()

        self.assertEqual(id(cfg1), id(cfg2))

    def test_006_scaffold(self):
        cmd = Command()

        # folder not exists, create
        answers = [self.output, "Y", "11"]
        with unittest.mock.patch("builtins.input", side_effect=answers):
            cmd.scaffold(None)

        self.assertTrue(os.path.exists(os.path.join(self.output, "defaults.yaml")))

        # files exists, ask for dst again, then we give non-exist path and 'Q' to quit
        answers = ["", "Q"]

        with unittest.mock.patch("builtins.input", side_effect=answers):
            try:
                with unittest.mock.patch(
                    "sys.exit", lambda *args: early_jump("user pressed Q")
                ):
                    cmd.scaffold(self.output)
                self.assertTrue(False, "sys exit not triggered")
            except SystemExit:
                self.assertTrue(True)

        # path exists and the folder is empty
        shutil.rmtree(self.output, ignore_errors=True)
        os.makedirs(self.output, exist_ok=True)
        answers = ["31"]
        with unittest.mock.patch("builtins.input", side_effect=answers):
            cmd.scaffold(self.output)

        self.assertTrue(os.path.exists(os.path.join(self.output, "defaults.yaml")))

        # non exist index
        shutil.rmtree(self.output, ignore_errors=True)
        os.makedirs(self.output, exist_ok=True)
        with unittest.mock.patch("builtins.input", side_effect=["99"]):
            cmd.scaffold(self.output)

        # chose logging
        shutil.rmtree(self.output, ignore_errors=True)
        os.makedirs(self.output, exist_ok=True)
        with unittest.mock.patch("builtins.input", side_effect=["0"]):
            cmd.scaffold(self.output)

    def test_007_cmd_build(self):
        cmd = Command()

        os.makedirs(self.output)
        with unittest.mock.patch("builtins.input", side_effect=["31"]):
            cmd.scaffold(self.output)

        # normal run
        cmd.build(self.output)
        cfg = cfg4py.init(self.output, strict=True)
        self.assertEqual(cfg.postgres.dsn, "dbname=test user=postgres password=secret")

        # path not exists run
        shutil.rmtree(self.output)
        try:
            with unittest.mock.patch(
                "sys.exit", lambda *args: early_jump("Path not exists")
            ):
                cmd.build(self.output)
            self.assertTrue(False, "sys.exit not triggered")
        except SystemExit:
            self.assertTrue(True)

        # no files in self.output
        os.makedirs(self.output, exist_ok=True)
        try:
            with unittest.mock.patch(
                "sys.exit", lambda *args: early_jump("no files in folder")
            ):
                cmd.build(self.output)
            self.assertTrue(False, "sys.exit not triggered")
        except SystemExit:
            self.assertTrue(True)

    def test_008_enable_logging(self):
        os.makedirs(self.output, 0o777, exist_ok=True)
        cfg4py.enable_logging(
            log_file=os.path.join(self.output, "cfg4py.log"),
            file_size=1e-6,
            file_count=3,
        )
        logger.info("%s", ["test log"])

    def test_009_set_server_role(self):
        cmd = Command()
        cmd.set_server_role()

    def test_010_read_envar(self):
        tmpdir = tempfile.gettempdir()
        cfg4dir = os.path.join(tmpdir, "cfg4py_test")
        config_file = os.path.join(cfg4dir, "defaults.yaml")
        os.makedirs(cfg4dir, 0o777, exist_ok=True)
        os.environ["account"] = "aaron"
        content = ["account: ${account}", "\n"]

        with open(config_file, "w+") as f:
            f.writelines("\n".join(content))

        cfg = cfg4py.init(cfg4dir, strict=True)
        self.assertEqual(cfg.account, "aaron")

    def test_011_non_strict_mode(self):
        os.environ[cfg4py.envar] = ""
        cfg = cfg4py.init(self.resource_path, strict=False)
        self.assertEqual(cfg.services.redis.host, "localhost")
